@@ -557,7 +557,7 @@ _wnck_get_utf8_list (Window   xwindow,
    * property is nul-separated
    */
   i = 0;
-  n_strings = 1;
+  n_strings = 0;
   while (i < nitems)
     {
       if (val[i] == '\0')
@@ -565,6 +565,9 @@ _wnck_get_utf8_list (Window   xwindow,
       ++i;
     }
 
+  if (val[nitems - 1] != '\0')
+    ++n_strings;
+
   /* we're guaranteed that val has a nul on the end
    * by XGetWindowProperty
    */
@@ -1854,7 +1857,7 @@ void
 _wnck_icon_cache_set_want_fallback (WnckIconCache *icon_cache,
                                     gboolean       setting)
 {
-  icon_cache->want_fallback = TRUE;
+  icon_cache->want_fallback = setting;
 }
 
 gboolean
@@ -2311,7 +2314,7 @@ _wnck_set_desktop_layout (Screen *xscreen,
                    RootWindowOfScreen (xscreen),
 		   _wnck_atom_get ("_NET_DESKTOP_LAYOUT"),
 		   XA_CARDINAL, 32, PropModeReplace,
-		   (guchar *)&data, sizeof (data) / 4);
+		   (guchar *)&data, 4);
 
   _wnck_error_trap_pop ();
 }
